// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package user

import (
	"fmt"
	"os"
	"runtime"
	"strings"
	"unsafe"
)

/*
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <stdlib.h>

static int mygetpwuid_r(int uid, struct passwd *pwd,
	char *buf, size_t buflen, struct passwd **result) {
 return getpwuid_r(uid, pwd, buf, buflen, result);
}

static int mygetgrgid_r(int gid, struct group *grp,
	char *buf, size_t buflen, struct group **result) {
 return getgrgid_r(gid, grp, buf, buflen, result);
}

static int mygetgrouplist(const char *user, gid_t group, gid_t *groups,
	int *ngroups) {
 return getgrouplist(user, group, (gid_t *)groups, ngroups);
}

static inline gid_t group_at(int i, gid_t *groups) {
 return groups[i];
}
static inline char **next_member(char **members) { return members + 1; }
*/
import "C"

// Lookup looks up a user by username. If the user cannot be found,
// the returned error is of type UnknownUserError.
func Lookup(username string) (*User, os.Error) {
	u, err := lookupUser(username, buildUser)
	if err != nil {
		return nil, err
	}
	return u, nil
}

// LookupId looks up a user by userid. If the user cannot be found,
// the returned error is of type UnknownUserIdError.
func LookupId(uid int) (*User, os.Error) {
	u, err := lookupUserId(uid, buildUser)
	if err != nil {
		return nil, err
	}
	return u, nil
}

// In indicates whether the user is a member of the given group.
func (u *User) In(g *Group) (bool, os.Error) {
	if u.Gid == g.Gid {
		return true, nil
	}

	nameC := C.CString(u.Username)
	defer C.free(unsafe.Pointer(nameC))
	groupC := C.gid_t(u.Gid)
	ngroupsC := C.int(0)

	C.mygetgrouplist(nameC, groupC, nil, &ngroupsC)
	ngroups := int(ngroupsC)

	groups := C.malloc(C.size_t(int(unsafe.Sizeof(groupC)) * ngroups))
	defer C.free(groups)

	rv := C.mygetgrouplist(nameC, groupC, (*C.gid_t)(groups), &ngroupsC)
	if rv == -1 {
		return false, fmt.Errorf("user: membership of %s in %s: %s", u.Username, g.Name, os.Errno(rv))
	}

	ngroups = int(ngroupsC)
	for i := 0; i < ngroups; i++ {
		gid := C.group_at(C.int(i), (*C.gid_t)(groups))
		if g.Gid == int(gid) {
			return true, nil
		}
	}
	return false, nil
}

// LookupGroup looks up a group by groupname. If the group cannot
// be found, the returned error is of type UnknownGroupError.
func LookupGroup(groupname string) (*Group, os.Error) {
	g, err := lookupGroup(groupname, buildGroup)
	if err != nil {
		return nil, err
	}
	return g, nil
}

// LookupGroupId looks up a group by groupid. If the group cannot
// be found, the returned error is of type UnknownGroupIdError.
func LookupGroupId(gid int) (*Group, os.Error) {
	g, err := lookupGroupId(gid, buildGroup)
	if err != nil {
		return nil, err
	}
	return g, nil
}

// Members returns the list of members of the group.
func (g *Group) Members() ([]string, os.Error) {
	var members []string

	_, err := lookupGroupId(g.Gid, func(grp *C.struct_group) *Group {
		cmem := grp.gr_mem
		for *cmem != nil {
			members = append(members, C.GoString(*cmem))
			cmem = C.next_member(cmem)
		}
		return g
	})
	if err != nil {
		return nil, err
	}

	return members, nil
}

const (
	userBuffer = iota
	groupBuffer
)

func lookupUser(username string, f func(*C.struct_passwd) *User) (*User, os.Error) {
	var pwd C.struct_passwd
	var result *C.struct_passwd

	buf, bufSize, err := allocBuffer(userBuffer)
	if err != nil {
		return nil, err
	}
	defer C.free(buf)

	nameC := C.CString(username)
	defer C.free(unsafe.Pointer(nameC))
	rv := C.getpwnam_r(nameC,
		&pwd,
		(*C.char)(buf),
		C.size_t(bufSize),
		&result)
	if rv != 0 {
		return nil, fmt.Errorf("user: lookup username %s: %s", username, os.Errno(rv))
	}
	if result == nil {
		return nil, UnknownUserError(username)
	}
	u := f(&pwd)
	return u, nil
}

func lookupUserId(uid int, f func(*C.struct_passwd) *User) (*User, os.Error) {
	var pwd C.struct_passwd
	var result *C.struct_passwd

	buf, bufSize, err := allocBuffer(userBuffer)
	if err != nil {
		return nil, err
	}
	defer C.free(buf)

	// mygetpwuid_r is a wrapper around getpwuid_r to
	// to avoid using uid_t because C.uid_t(uid) for
	// unknown reasons doesn't work on linux.
	rv := C.mygetpwuid_r(C.int(uid),
		&pwd,
		(*C.char)(buf),
		C.size_t(bufSize),
		&result)
	if rv != 0 {
		return nil, fmt.Errorf("user: lookup userid %d: %s", uid, os.Errno(rv))
	}
	if result == nil {
		return nil, UnknownUserIdError(uid)
	}
	u := f(&pwd)
	return u, nil
}

func buildUser(pwd *C.struct_passwd) *User {
	u := &User{
		Uid:      int(pwd.pw_uid),
		Gid:      int(pwd.pw_gid),
		Username: C.GoString(pwd.pw_name),
		Name:     C.GoString(pwd.pw_gecos),
		HomeDir:  C.GoString(pwd.pw_dir),
	}
	// The pw_gecos field isn't quite standardized.  Some docs
	// say: "It is expected to be a comma separated list of
	// personal data where the first item is the full name of the
	// user."
	if i := strings.Index(u.Name, ","); i >= 0 {
		u.Name = u.Name[:i]
	}
	return u
}

func lookupGroup(groupname string, f func(*C.struct_group) *Group) (*Group, os.Error) {
	var grp C.struct_group
	var result *C.struct_group

	buf, bufSize, err := allocBuffer(groupBuffer)
	if err != nil {
		return nil, err
	}
	defer C.free(buf)

	nameC := C.CString(groupname)
	defer C.free(unsafe.Pointer(nameC))
	rv := C.getgrnam_r(nameC,
		&grp,
		(*C.char)(buf),
		C.size_t(bufSize),
		&result)
	if rv != 0 {
		return nil, fmt.Errorf("group: lookup groupname %s: %s", groupname, os.Errno(rv))
	}
	if result == nil {
		return nil, UnknownGroupError(groupname)
	}
	g := f(&grp)
	return g, nil
}

func lookupGroupId(gid int, f func(*C.struct_group) *Group) (*Group, os.Error) {
	var grp C.struct_group
	var result *C.struct_group

	buf, bufSize, err := allocBuffer(groupBuffer)
	if err != nil {
		return nil, err
	}
	defer C.free(buf)

	// mygetgrgid_r is a wrapper around getgrgid_r to
	// to avoid using gid_t because C.gid_t(gid) for
	// unknown reasons doesn't work on linux.
	rv := C.mygetgrgid_r(C.int(gid),
		&grp,
		(*C.char)(buf),
		C.size_t(bufSize),
		&result)
	if rv != 0 {
		return nil, fmt.Errorf("group: lookup groupid %d: %s", gid, os.Errno(rv))
	}
	if result == nil {
		return nil, UnknownGroupIdError(gid)
	}
	g := f(&grp)
	return g, nil
}

func buildGroup(grp *C.struct_group) *Group {
	g := &Group{
		Gid:  int(grp.gr_gid),
		Name: C.GoString(grp.gr_name),
	}
	return g
}

func allocBuffer(bufType int) (unsafe.Pointer, C.long, os.Error) {
	var bufSize C.long
	if runtime.GOOS == "freebsd" {
		// FreeBSD doesn't have _SC_GETPW_R_SIZE_MAX
		// or SC_GETGR_R_SIZE_MAX and just returns -1.
		// So just use the same size that Linux returns
		bufSize = 1024
	} else {
		var size C.int
		var constName string
		switch bufType {
		case userBuffer:
			size = C._SC_GETPW_R_SIZE_MAX
			constName = "_SC_GETPW_R_SIZE_MAX"
		case groupBuffer:
			size = C._SC_GETGR_R_SIZE_MAX
			constName = "_SC_GETGR_R_SIZE_MAX"
		}
		bufSize = C.sysconf(size)
		if bufSize <= 0 || bufSize > 1<<20 {
			return nil, bufSize, fmt.Errorf("user: unreasonable %s of %d", constName, bufSize)
		}
	}
	return C.malloc(C.size_t(bufSize)), bufSize, nil
}
